﻿using MySql.Data.MySqlClient;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;

namespace MontoyaTech.MySqlPlus
{
    /// <summary>
    /// A set of extensions to help work with MySqlCommands.
    /// </summary>
    public static class MySqlCommandExtensions
    {
        /// <summary>
        /// Setups this MySqlCommand to insert a row in the db.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="command"></param>
        /// <param name="row"></param>
        /// <exception cref="Exception"></exception>
        public static void Insert<T>(this MySqlCommand command, T row)
        {
            //Get the type of T
            var type = typeof(T);

            //Get the row information.
            var rowAttribute = type.GetCustomAttribute<MySqlRow>();

            //Get all the fields.
            var fields = type.GetFields();

            if (fields == null || fields.Length == 0)
                throw new Exception("Found no public fields on given row.");

            //Start building the query.
            var builder = new StringBuilder();

            //Write the insert section.
            builder.Append($"INSERT INTO `{(rowAttribute == null || string.IsNullOrWhiteSpace(rowAttribute.Name) ? type.Name : rowAttribute.Name)}` ");

            //Write the set values section.
            builder.Append("SET ");

            bool seperate = false;
            for (int i = 0; i < fields.Length; i++)
            {
                var column = fields[i].GetCustomAttribute<MySqlColumn>();

                if (column == null)
                    continue;

                //Skip id columns because they are auto incremented.
                if (column.Id)
                    continue;

                if (seperate)
                    builder.Append(", ");

                builder.Append($"`{(string.IsNullOrWhiteSpace(column.Name) ? fields[i].Name : column.Name)}` = @{i}");

                command.Parameters.AddWithValue($"@{i}", fields[i].GetValue(row));

                seperate = true;
            }

            //Set the command text.
            command.CommandText = builder.ToString();
        }

        /// <summary>
        /// Setups this MySqlCommand to update a row in the db.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="command"></param>
        /// <param name="row"></param>
        /// <exception cref="Exception"></exception>
        public static void Update<T>(this MySqlCommand command, T row)
        {
            //Get the type of T
            var type = typeof(T);

            //Get the row information.
            var rowAttribute = type.GetCustomAttribute<MySqlRow>();

            //Get all the fields.
            var fields = type.GetFields();

            if (fields == null || fields.Length == 0)
                throw new Exception("Found no public fields on given row.");

            //Start building the query.
            var builder = new StringBuilder();

            //Write the update section
            builder.Append($"UPDATE `{(rowAttribute == null || string.IsNullOrWhiteSpace(rowAttribute.Name) ? type.Name : rowAttribute.Name)}` ");

            //Write the set values section
            builder.Append("SET ");

            FieldInfo idField = null;
            MySqlColumn idColumn = null;
            bool seperate = false;
            for (int i = 0; i < fields.Length; i++)
            {
                var column = fields[i].GetCustomAttribute<MySqlColumn>();

                if (column == null)
                    continue;

                //Skip id columns because they are auto incremented.
                if (column.Id)
                {
                    idField = fields[i];
                    idColumn = column;
                    continue;
                }

                if (seperate)
                    builder.Append(", ");

                builder.Append($"`{(string.IsNullOrWhiteSpace(column.Name) ? fields[i].Name : column.Name)}` = @{i}");

                command.Parameters.AddWithValue($"@{i}", fields[i].GetValue(row));

                seperate = true;
            }

            //Make sure we have an id column.
            if (idField == null)
                throw new Exception("Given row does not contain an id column");

            //Write the where clause
            builder.Append($" WHERE `{(string.IsNullOrWhiteSpace(idColumn.Name) ? idField.Name : idColumn.Name)}`=@id LIMIT 1");

            command.Parameters.AddWithValue("@id", idField.GetValue(row));

            //Set the command text.
            command.CommandText = builder.ToString();
        }

        /// <summary>
        /// Setups this MySqlCommand to get a row from the db by it's id.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="command"></param>
        /// <param name="id"></param>
        /// <exception cref="Exception"></exception>
        public static void Get<T>(this MySqlCommand command, ulong id)
        {
            //Get the type of T
            var type = typeof(T);

            //Get the row information.
            var rowAttribute = type.GetCustomAttribute<MySqlRow>();

            //Get all the fields.
            var fields = type.GetFields();

            if (fields == null || fields.Length == 0)
                throw new Exception("Found no public fields on given row.");

            //Get the id field
            if (!type.GetMySqlId(out FieldInfo idField, out MySqlColumn idColumn))
                throw new Exception("Failed to find the id column on the given row.");

            //Set the command text.
            command.CommandText = $"SELECT * FROM `{(rowAttribute == null || string.IsNullOrWhiteSpace(rowAttribute.Name) ? type.Name : rowAttribute.Name)}` WHERE `{(string.IsNullOrWhiteSpace(idColumn.Name) ? idField.Name : idColumn.Name)}`=@id LIMIT 1";
            
            //Add the id parameter.
            command.Parameters.AddWithValue("@id", id);
        }

        /// <summary>
        /// Setups this MySqlCommand to get all the rows of a given type from the db.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="command"></param>
        public static void GetAll<T>(this MySqlCommand command)
        {
            //Get the type of T
            var type = typeof(T);

            //Get the row information.
            var rowAttribute = type.GetCustomAttribute<MySqlRow>();

            //Set the command text.
            command.CommandText = $"SELECT * FROM `{(rowAttribute == null || string.IsNullOrWhiteSpace(rowAttribute.Name) ? type.Name : rowAttribute.Name)}`";
        }

        /// <summary>
        /// Setups this MySqlCommand to delete a row from the db by it's id from an instance of the row.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="command"></param>
        /// <param name="row"></param>
        /// <exception cref="Exception"></exception>
        public static void Delete<T>(this MySqlCommand command, T row)
        {
            //Get the type of T
            var type = typeof(T);

            //Get the row information.
            var rowAttribute = type.GetCustomAttribute<MySqlRow>();

            //Start building the query.
            var builder = new StringBuilder();

            //Write the delete from section
            builder.Append($"DELETE FROM `{(rowAttribute == null || string.IsNullOrWhiteSpace(rowAttribute.Name) ? type.Name : rowAttribute.Name)}` ");

            //Get the id the column and field info
            if (!row.GetMySqlId(out FieldInfo idField, out MySqlColumn idColumn))
                throw new Exception("Failed to find Id column on row.");

            //Write the where clause
            builder.Append($"WHERE `{(string.IsNullOrWhiteSpace(idColumn.Name) ? idField.Name : idColumn.Name)}`=@id LIMIT 1");

            //Add the id parameter.
            command.Parameters.AddWithValue("@id", idField.GetValue(row));

            //Set the command text.
            command.CommandText = builder.ToString();
        }

        /// <summary>
        /// Setups this MySqlCommand to delete a row from the db by it's id.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="command"></param>
        /// <param name="id"></param>
        public static void Delete<T>(this MySqlCommand command, ulong id)
        {
            //Get the type of T
            var type = typeof(T);

            //Get the row information.
            var rowAttribute = type.GetCustomAttribute<MySqlRow>();

            //Start building the query.
            var builder = new StringBuilder();

            //Write the delete from section
            builder.Append($"DELETE FROM `{(rowAttribute == null || string.IsNullOrWhiteSpace(rowAttribute.Name) ? type.Name : rowAttribute.Name)}` ");

            //Get the id the column and field info
            if (!type.GetMySqlId(out FieldInfo idField, out MySqlColumn idColumn))
                throw new Exception("Failed to find Id column on row.");

            //Write the where clause
            builder.Append($"WHERE `{(string.IsNullOrWhiteSpace(idColumn.Name) ? idField.Name : idColumn.Name)}`=@id LIMIT 1");

            //Add the id parameter.
            command.Parameters.AddWithValue("@id", id);

            //Set the command text.
            command.CommandText = builder.ToString();
        }
    }
}
